-module(glyph@internal@network@gateway).
-compile([no_auto_import, nowarn_unused_vars, nowarn_unused_function, nowarn_nomatch]).

-export([start_gateway_actor/4]).
-export_type([msg/0, gateway_error/0, actor_state/0]).

-type msg() :: close | heartbeat | identify | resume.

-type gateway_error() :: {json_error, gleam@json:decode_error()} |
    {dyn_error, gleam@dynamic:decode_error()} |
    {dyn_errors, list(gleam@dynamic:decode_error())} |
    {actor_error, gleam@otp@actor:start_error()} |
    invalid_session_error.

-type actor_state() :: {actor_state,
        gleam@erlang@process:subject(msg()),
        binary(),
        glyph@models@discord:event_handler(),
        carpenter@table:set(binary(), binary()),
        boolean(),
        integer(),
        gleam@option:option(integer()),
        boolean()}.

-spec fallback_event() -> glyph@models@discord:gateway_event().
fallback_event() ->
    {gateway_event, -1, gleam@dynamic:from(<<""/utf8>>), none, none}.

-spec fallback_hello() -> glyph@models@discord:hello_event().
fallback_hello() ->
    {hello_event, -1}.

-spec heartbeat_json(gleam@option:option(integer())) -> binary().
heartbeat_json(Seq) ->
    _pipe = gleam@json:object(
        [{<<"op"/utf8>>, gleam@json:int(1)},
            {<<"d"/utf8>>, gleam@json:nullable(Seq, fun gleam@json:int/1)}]
    ),
    gleam@json:to_string(_pipe).

-spec identify_json(binary(), integer(), binary()) -> binary().
identify_json(Token, Intents, Os) ->
    Properties = gleam@json:object(
        [{<<"os"/utf8>>, gleam@json:string(Os)},
            {<<"browser"/utf8>>, gleam@json:string(<<"glyph"/utf8>>)},
            {<<"device"/utf8>>, gleam@json:string(<<"glyph"/utf8>>)}]
    ),
    _pipe = gleam@json:object(
        [{<<"op"/utf8>>, gleam@json:int(2)},
            {<<"d"/utf8>>,
                gleam@json:object(
                    [{<<"intents"/utf8>>, gleam@json:int(Intents)},
                        {<<"token"/utf8>>, gleam@json:string(Token)},
                        {<<"properties"/utf8>>, Properties}]
                )}]
    ),
    gleam@json:to_string(_pipe).

-spec resume_json(binary(), binary(), integer()) -> binary().
resume_json(Token, Session_id, Seq) ->
    _pipe = gleam@json:object(
        [{<<"op"/utf8>>, gleam@json:int(6)},
            {<<"d"/utf8>>,
                gleam@json:object(
                    [{<<"token"/utf8>>, gleam@json:string(Token)},
                        {<<"session_id"/utf8>>, gleam@json:string(Session_id)},
                        {<<"seq"/utf8>>, gleam@json:int(Seq)}]
                )}]
    ),
    gleam@json:to_string(_pipe).

-spec jitter(float()) -> integer().
jitter(Heartbeat_interval) ->
    Gen_jitter = prng@random:float(+0.0, 1.0),
    _pipe = (prng@random:random_sample(Gen_jitter) * Heartbeat_interval),
    gleam@float:truncate(_pipe).

-spec handle_gateway_recv(
    binary(),
    actor_state(),
    stratus:connection(),
    glyph@models@discord:bot_client()
) -> actor_state().
handle_gateway_recv(Msg, State, Conn, Bot) ->
    Event = begin
        _pipe = gleam@json:decode(
            Msg,
            glyph@internal@decoders:gateway_event_decoder()
        ),
        gleam@result:unwrap(_pipe, fallback_event())
    end,
    case erlang:element(2, Event) of
        0 ->
            logging:log(debug, <<"Received Dispatch event from gateway"/utf8>>),
            Event_type = gleam@option:unwrap(
                erlang:element(5, Event),
                <<""/utf8>>
            ),
            case Event_type of
                <<"READY"/utf8>> ->
                    Ready = glyph@internal@decoders:gateway_ready_decoder(
                        erlang:element(3, Event)
                    ),
                    case Ready of
                        {ok, Ev} ->
                            _pipe@1 = erlang:element(5, State),
                            carpenter@table:insert(
                                _pipe@1,
                                <<"resume_gateway_url"/utf8>>,
                                erlang:element(6, Ev)
                            ),
                            erlang:setelement(
                                3,
                                State,
                                erlang:element(2, erlang:element(3, Ev))
                            );

                        {error, _} ->
                            logging:log(
                                error,
                                <<"Error parsing Ready event"/utf8>>
                            ),
                            State
                    end;

                <<"MESSAGE_CREATE"/utf8>> ->
                    logging:log(debug, <<"Handling MESSAGE_CREATE event"/utf8>>),
                    Message = glyph@internal@decoders:message_decoder(
                        erlang:element(3, Event)
                    ),
                    case Message of
                        {ok, Msg@1} ->
                            case erlang:element(2, erlang:element(4, Msg@1)) =:= erlang:element(
                                3,
                                State
                            ) of
                                true ->
                                    logging:log(
                                        debug,
                                        <<"I think that I am the author"/utf8>>
                                    ),
                                    State;

                                false ->
                                    logging:log(
                                        debug,
                                        <<"Invoking on_message_create"/utf8>>
                                    ),
                                    _ = (erlang:element(
                                        2,
                                        erlang:element(4, State)
                                    ))(Bot, Msg@1),
                                    State
                            end;

                        {error, _} ->
                            logging:log(
                                error,
                                <<"Error parsing MESSAGE_CREATE event"/utf8>>
                            ),
                            State
                    end,
                    State;

                <<""/utf8>> ->
                    logging:log(
                        warning,
                        <<"Received Dispatch event with no type"/utf8>>
                    ),
                    State;

                U ->
                    logging:log(
                        warning,
                        <<"Received Dispatch event with unsupported type: "/utf8,
                            U/binary>>
                    ),
                    State
            end;

        1 ->
            logging:log(
                debug,
                <<"Received Heartbeat request from gateway"/utf8>>
            ),
            _ = stratus:send_text_message(
                Conn,
                heartbeat_json(erlang:element(8, State))
            ),
            State;

        7 ->
            logging:log(
                debug,
                <<"Received Reconnect request from gateway"/utf8>>
            ),
            _pipe@2 = erlang:element(5, State),
            carpenter@table:insert(
                _pipe@2,
                <<"should_resume"/utf8>>,
                <<"true"/utf8>>
            ),
            gleam@erlang@process:send(erlang:element(2, State), close),
            State;

        9 ->
            logging:log(
                warning,
                <<"Received Invalid Session from gateway"/utf8>>
            ),
            State;

        10 ->
            case erlang:element(6, State) of
                false ->
                    logging:log(debug, <<"Received Hello from gateway"/utf8>>),
                    Hello = begin
                        _pipe@3 = glyph@internal@decoders:gateway_hello_decoder(
                            erlang:element(3, Event)
                        ),
                        gleam@result:unwrap(_pipe@3, fallback_hello())
                    end,
                    Jit = jitter(gleam@int:to_float(erlang:element(2, Hello))),
                    logging:log(
                        debug,
                        <<"Jitter value: "/utf8,
                            (gleam@int:to_string(Jit))/binary>>
                    ),
                    gleam@erlang@process:send_after(
                        erlang:element(2, State),
                        Jit,
                        heartbeat
                    ),
                    Should_resume = glyph@internal@cache:should_resume(
                        erlang:element(5, State),
                        false
                    ),
                    case Should_resume of
                        true ->
                            gleam@erlang@process:send(
                                erlang:element(2, State),
                                resume
                            );

                        false ->
                            gleam@erlang@process:send(
                                erlang:element(2, State),
                                identify
                            )
                    end,
                    erlang:setelement(
                        6,
                        erlang:setelement(7, State, erlang:element(2, Hello)),
                        true
                    );

                true ->
                    State
            end;

        11 ->
            logging:log(debug, <<"Received Heartbeat ACK from gateway"/utf8>>),
            State;

        4000 ->
            logging:log(
                error,
                <<"An unknown error occurred. Attempting reconnect."/utf8>>
            ),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"should_resume"/utf8>>,
                <<"true"/utf8>>
            ),
            gleam@erlang@process:send(erlang:element(2, State), close),
            State;

        4001 ->
            logging:log(
                error,
                <<"You sent an invalid opcode or an invalid payload for an opcode."/utf8>>
            ),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"should_resume"/utf8>>,
                <<"true"/utf8>>
            ),
            gleam@erlang@process:send(erlang:element(2, State), close),
            State;

        4002 ->
            logging:log(error, <<"You sent an invalid payload."/utf8>>),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"should_resume"/utf8>>,
                <<"true"/utf8>>
            ),
            gleam@erlang@process:send(erlang:element(2, State), close),
            State;

        4003 ->
            logging:log(
                error,
                <<"You sent a payload prior to identifying."/utf8>>
            ),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"should_resume"/utf8>>,
                <<"false"/utf8>>
            ),
            gleam@erlang@process:send(erlang:element(2, State), close),
            State;

        4004 ->
            logging:log(
                error,
                <<"The account token sent with your identify payload is incorrect."/utf8>>
            ),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"should_resume"/utf8>>,
                <<"false"/utf8>>
            ),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"invalid_session"/utf8>>,
                <<"true"/utf8>>
            ),
            gleam@erlang@process:send(erlang:element(2, State), close),
            State;

        4005 ->
            logging:log(
                error,
                <<"You send more than one identify payload."/utf8>>
            ),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"should_resume"/utf8>>,
                <<"true"/utf8>>
            ),
            gleam@erlang@process:send(erlang:element(2, State), close),
            State;

        4007 ->
            logging:log(
                error,
                <<"Invalid sequence sent when resuming the session."/utf8>>
            ),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"should_resume"/utf8>>,
                <<"false"/utf8>>
            ),
            gleam@erlang@process:send(erlang:element(2, State), close),
            State;

        4008 ->
            logging:log(
                error,
                <<"You have been rate limited for sending too many requests."/utf8>>
            ),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"should_resume"/utf8>>,
                <<"true"/utf8>>
            ),
            gleam@erlang@process:send(erlang:element(2, State), close),
            State;

        4009 ->
            logging:log(error, <<"Session timed out."/utf8>>),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"should_resume"/utf8>>,
                <<"false"/utf8>>
            ),
            gleam@erlang@process:send(erlang:element(2, State), close),
            State;

        4010 ->
            logging:log(
                error,
                <<"You sent an invalid shard when identifying."/utf8>>
            ),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"should_resume"/utf8>>,
                <<"false"/utf8>>
            ),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"invalid_session"/utf8>>,
                <<"true"/utf8>>
            ),
            gleam@erlang@process:send(erlang:element(2, State), close),
            State;

        4011 ->
            logging:log(
                error,
                <<"The session would have handled too many guilds - shard your connection to connect."/utf8>>
            ),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"should_resume"/utf8>>,
                <<"false"/utf8>>
            ),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"invalid_session"/utf8>>,
                <<"true"/utf8>>
            ),
            gleam@erlang@process:send(erlang:element(2, State), close),
            State;

        4012 ->
            logging:log(
                error,
                <<"You sent an invalid version for the gateway."/utf8>>
            ),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"should_resume"/utf8>>,
                <<"false"/utf8>>
            ),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"invalid_session"/utf8>>,
                <<"true"/utf8>>
            ),
            gleam@erlang@process:send(erlang:element(2, State), close),
            State;

        4013 ->
            logging:log(
                error,
                <<"You sent an invalid intent for a Gateway Intent. You may have incorrectly calculated the bitwise value."/utf8>>
            ),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"should_resume"/utf8>>,
                <<"false"/utf8>>
            ),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"invalid_session"/utf8>>,
                <<"true"/utf8>>
            ),
            gleam@erlang@process:send(erlang:element(2, State), close),
            State;

        4014 ->
            logging:log(
                error,
                <<"You sent a disallowed intent for a Gateway Intent. You may have tried to specify an intent that you have not enabled or are not approved for."/utf8>>
            ),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"should_resume"/utf8>>,
                <<"false"/utf8>>
            ),
            carpenter@table:insert(
                erlang:element(5, State),
                <<"invalid_session"/utf8>>,
                <<"true"/utf8>>
            ),
            gleam@erlang@process:send(erlang:element(2, State), close),
            State;

        N ->
            logging:log(
                warning,
                <<"Received unexpected op code "/utf8,
                    (gleam@int:to_string(N))/binary>>
            ),
            State
    end.

-spec determine_url(carpenter@table:set(binary(), binary()), binary()) -> gleam@http@request:request(binary()).
determine_url(Session_cache, Url) ->
    Resume_url = glyph@internal@cache:resume_gateway_url(Session_cache, Url),
    Should_resume = glyph@internal@cache:should_resume(Session_cache, false),
    Base_url = case Should_resume of
        true ->
            gleam@string:replace(Resume_url, <<"wss"/utf8>>, <<"https"/utf8>>);

        false ->
            gleam@string:replace(Url, <<"wss"/utf8>>, <<"https"/utf8>>)
    end,
    _assert_subject = gleam@http@request:to(
        <<<<<<Base_url/binary, "/?v="/utf8>>/binary, (<<"10"/utf8>>)/binary>>/binary,
            "&encoding=json"/utf8>>
    ),
    {ok, Init_req} = case _assert_subject of
        {ok, _} -> _assert_subject;
        _assert_fail ->
            erlang:error(#{gleam_error => let_assert,
                        message => <<"Assertion pattern match failed"/utf8>>,
                        value => _assert_fail,
                        module => <<"glyph/internal/network/gateway"/utf8>>,
                        function => <<"determine_url"/utf8>>,
                        line => 472})
    end,
    Init_req.

-spec start_gateway_actor(
    gleam@erlang@process:subject(gleam@erlang@process:subject(msg())),
    glyph@models@discord:bot_client(),
    binary(),
    carpenter@table:set(binary(), binary())
) -> {ok, gleam@erlang@process:subject(stratus:internal_message(msg()))} |
    {error, gleam@otp@actor:start_error()}.
start_gateway_actor(Supervisor_subj, Bot, Url, Session_cache) ->
    Init_req = determine_url(Session_cache, Url),
    Builder = begin
        _pipe@1 = stratus:websocket(
            Init_req,
            fun() ->
                Actor_subj = gleam@erlang@process:new_subject(),
                gleam@erlang@process:send(Supervisor_subj, Actor_subj),
                Selector = begin
                    _pipe = gleam_erlang_ffi:new_selector(),
                    gleam@erlang@process:selecting(
                        _pipe,
                        Actor_subj,
                        fun gleam@function:identity/1
                    )
                end,
                {{actor_state,
                        Actor_subj,
                        <<""/utf8>>,
                        erlang:element(7, Bot),
                        Session_cache,
                        false,
                        0,
                        none,
                        glyph@internal@cache:invalid_session(
                            Session_cache,
                            false
                        )},
                    {some, Selector}}
            end,
            fun(Msg, State, Conn) -> case {Msg, erlang:element(9, State)} of
                    {_, true} ->
                        logging:log(
                            error,
                            <<"Invalid session, refusing to connect."/utf8>>
                        ),
                        gleam@otp@actor:continue(State);

                    {{text, Msg@1}, _} ->
                        logging:log(debug, <<"RECV: "/utf8, Msg@1/binary>>),
                        New_state = handle_gateway_recv(Msg@1, State, Conn, Bot),
                        gleam@otp@actor:continue(New_state);

                    {{user, heartbeat}, _} ->
                        logging:log(debug, <<"Sending heartbeat"/utf8>>),
                        _ = stratus:send_text_message(
                            Conn,
                            heartbeat_json(erlang:element(8, State))
                        ),
                        gleam@erlang@process:send_after(
                            erlang:element(2, State),
                            erlang:element(7, State),
                            heartbeat
                        ),
                        gleam@otp@actor:continue(State);

                    {{user, identify}, _} ->
                        logging:log(debug, <<"Sending Identify"/utf8>>),
                        _ = stratus:send_text_message(
                            Conn,
                            identify_json(
                                erlang:element(3, Bot),
                                erlang:element(6, Bot),
                                <<"linux"/utf8>>
                            )
                        ),
                        gleam@otp@actor:continue(State);

                    {{user, resume}, _} ->
                        logging:log(
                            debug,
                            <<"Attempting to resume session"/utf8>>
                        ),
                        Seq = glyph@internal@cache:seq(
                            erlang:element(5, State),
                            0
                        ),
                        Session_id = glyph@internal@cache:session_id(
                            erlang:element(5, State),
                            <<""/utf8>>
                        ),
                        _ = stratus:send_text_message(
                            Conn,
                            resume_json(erlang:element(3, Bot), Session_id, Seq)
                        ),
                        carpenter@table:insert(
                            erlang:element(5, State),
                            <<"should_resume"/utf8>>,
                            <<"false"/utf8>>
                        ),
                        gleam@otp@actor:continue(State);

                    {{user, close}, _} ->
                        logging:log(
                            info,
                            <<"Closing WebSocket connection"/utf8>>
                        ),
                        case stratus:close(Conn) of
                            {ok, _} ->
                                nil;

                            {error, E} ->
                                logging:log(
                                    error,
                                    <<"Error closing websocket: "/utf8,
                                        (gleam@string:inspect(E))/binary>>
                                )
                        end,
                        {stop, normal};

                    {_, _} ->
                        logging:log(warning, <<"Reached unexpected case"/utf8>>),
                        gleam@otp@actor:continue(State)
                end end
        ),
        stratus:on_close(
            _pipe@1,
            fun(_) ->
                logging:log(debug, <<"WebSocket process closing"/utf8>>)
            end
        )
    end,
    stratus:initialize(Builder).
